##########################################################################################################
#         R code for an univariate continuous-time model with frequentist estimation in ctsem            #
##########################################################################################################


## used R version: 3.6.1 (R Core Team, 2019)
# => get latest version: https://protect-us.mimecast.com/s/3yH6C2kg95ik6kB51t8WLZu?domain=cran.r-project.org/

## load/install packages
# used ctsem version: 2.9.6, Driver et al. (2019)
# => get latest ctsem version: install.packages( "ctsem" )
library( ctsem )
# used reshape2 version: 1.4.3
# => get latest reshape2 version: install.packages( "reshape2" )
library( reshape2 )


### Step 1: Set up the model

# number of processes (=latent variables)
F <- 1

# drift matrix
A <- matrix( "a", nrow=F, ncol=F )

# diffusion covariance matrix (parameter names consistent with output, not input)
Q <- matrix( "q", nrow=F, ncol=F )

# covariance matrix of process means
# (parameter names consistent with output, not input)
Sigma.mu <- matrix( "sigma2mu", nrow=F, ncol=F )

# within-person variance at first measurement occasion
Sigma.fw <- matrix( "sigma2fw", nrow=F, ncol=F )

# total mean at first measurement occasion
Mu.f <- matrix( "muplusmudev", nrow=F, ncol=F )

# continuous-time intercept
b <- matrix( "b", nrow=F, ncol=1 )

## elements for direct mapping of manifest variables onto one latent variable each (no measurement model)
# number of manifest variables
I <- F

# set all intercepts of manifest variables to 0
manifestmeans <- matrix( 0, nrow=I, ncol=1 )

# set measurement error covariance matrix to 0
manifestvar <- matrix( 0, nrow=I, ncol=I )

# loading matrix: each item loads on one latent variable
Lambda <- diag( F )

## set up model
m <- ctModel( n.latent = F,
              latentNames = c("theta"), # set names of latent variables
              n.manifest = I,
              manifestNames = c("y"),   # names of manifest variables as in data set
              DRIFT = A,
              DIFFUSION = Q,
              TRAITVAR = Sigma.mu,
              T0VAR = Sigma.fw,
              T0MEANS = Mu.f,
              CINT = b,
              LAMBDA = Lambda,
              MANIFESTMEANS = manifestmeans,
              MANIFESTVAR = manifestvar,
              Tpoints = 3,              # number of measurement occasions as in data set
              type = "omx" )            # frequentist estimation


### Step 2: Load and check the data

## load simulated example data
## (simulated data with J=10000 persons, P=3 measurement occasions, a=-0.40, q=0.40, sigma2mu=0.50, sigma2fw=0.50, mu=0, mudev=0)
( load( url( "https://protect-us.mimecast.com/s/OkcQC31j95hmxmyDqHzu1f1?domain=ndownloader.figshare.com" ) ) )
# D: data matrix in wide format:
#    rows:     measurement occasions
#    columns:  persons
#    cells:    values of persons at measurement occasions
# delta: time interval length between measurement occasions

# transform wide format D to long format L
L <- melt( D )
colnames( L ) <- c( "id", "p", "y" )
L <- L[ order(L$id,L$p), ]
rownames( L ) <- 1:nrow( L )

# create time variable
L$time <- rep( cumsum( c( 0, delta ) ), nrow( D ) )


### Step 3: Run the model
# approximate run time for this example: 4 s

# start time
start <- Sys.time()

# run the continuous-time model (frequentist estimation)
set.seed( 4321 ) # for reproducible standard errors
r <- ctFit( ctmodelobj = m, 
            dat = L,
            # stationary = "all", # uncomment to run model with stationary constraints for first measurement occasion
            dataform = "long" )

# run time
print( runtime <- Sys.time() - start )

## add additional parameters to the estimation
# retrieve mx object
mxobj <- r$mxobj
# process means matrix Mu
mxobj$Mu <- mxAlgebra( -invDRIFT %*% CINT, name="Mu" )
# deviation matrix of means at first measurement occasion from process means, Mudev
mxobj$Mudev <- mxAlgebra( T0MEANS - Mu, name="Mu" )

# run again with additional parameters
set.seed( 4321 ) # for reproducible standard errors
r2 <- mxRun( mxobj )


### Step 4: view results

smr <- summary( r )
print( smr$ctparameters )

## reported standard errors for variances in the used version of ctsem are flawed
# retrieve open mx standard errors instead
print( mxSE( DIFFUSION, model=r$mxobj ) )
print( mxSE( T0VAR, model=r$mxobj ) )
print( mxSE( TRAITVAR, model=r$mxobj ) )

## view additional parameters

# process mean mu
# estimate
print( mxEval( Mu, r2 ) )
# standard error
print( mxSE( Mu, model=r2 ) )

# deviation of mean at first measurement occasion
# estimate
print( mxEval( Mudev, r2 ) )
# standard error
print( mxSE( Mudev, model=r2 ) )
