##########################################################################################################
#          R code for an univariate continuous-time model with Bayesian estimation in ctsem              #
##########################################################################################################


## used R version: 3.6.1 (R Core Team, 2019)
# => get latest version: https://protect-us.mimecast.com/s/3yH6C2kg95ik6kB51t8WLZu?domain=cran.r-project.org

## load/install packages
# used ctsem version: 2.9.6, Driver et al. (2019)
# => get latest ctsem version: install.packages( "ctsem" )
library( ctsem )
# used reshape2 version: 1.4.3
# => get latest reshape2 version: install.packages( "reshape2" )
library( reshape2 )


### Step 1: Set up the model

# number of processes (=latent variables)
F <- 1

# drift matrix
A <- matrix( "a", nrow=F, ncol=F )

# diffusion covariance matrix (parameter names consistent with output, not input)
Q <- matrix( "sqrt.q", nrow=F, ncol=F )

# continuous-time intercept
b <- matrix( "b", nrow=F, ncol=1 )

## elements for direct mapping of manifest variables onto one latent variable each (no measurement model)
# number of manifest variables
I <- F

# set all intercepts of manifest variables to 0
manifestmeans <- matrix( 0, nrow=I, ncol=1 )

# set measurement error covariance matrix to 0
manifestvar <- matrix( 0, nrow=I, ncol=I )

# loading matrix: each item loads on one latent variable
Lambda <- diag( F )

## set up model
m <- ctModel( n.latent = F,
              latentNames = c("theta"), # set names of latent variables
              n.manifest = I,
              manifestNames = c("y"),   # names of manifest variables as in data set
              DRIFT = A,
              DIFFUSION = Q,
              CINT = b,
              LAMBDA = Lambda,
              MANIFESTMEANS = manifestmeans,
              MANIFESTVAR = manifestvar,
              type = "stanct" )         # Bayesian estimation
## set parameters that vary over persons
# set all parameters non-varying
m$pars$indvarying <- FALSE
# set continuous-time intercept varying
m$pars$indvarying[ m$pars$matrix %in% 'CINT' ] <- TRUE


### Step 2: Load and check the data

## load simulated example data
## (simulated data with J=10000 persons, P=3 measurement occasions, a=-0.40, q=0.40, sigma2mu=0.50, sigma2fw=0.50, mu=0, mudev=0)
( load( url( "https://protect-us.mimecast.com/s/OkcQC31j95hmxmyDqHzu1f1?domain=ndownloader.figshare.com" ) ) )
# D: data matrix in wide format:
#    rows:     measurement occasions
#    columns:  persons
#    cells:    values of persons at measurement occasions
# delta: time interval length between measurement occasions

# transform wide format D to long format L
L <- melt( D )
colnames( L ) <- c( "id", "p", "y" )
L <- L[ order(L$id,L$p), ]
rownames( L ) <- 1:nrow( L )

# create time variable
L$time <- rep( cumsum( c( 0, delta ) ), nrow( D ) )

# reduce data set for faster estimation, only for illustration purpose!
L <- L[1:300,]


### Step 3: Run the model

# start time
start <- Sys.time()

# run the continuous-time model (frequentist estimation)
# run time of this example: approximately 8 min
set.seed( 4321 ) # for reproducible results
r <- ctStanFit( datalong    = L,
                ctstanmodel = m,
                iter        = 100,  # low number of iterations, only for illustration purpose!
                chains      = 1,
                cores       = 1,
                stationary  = TRUE, # stationary constraints for first measurement occasion
                savescores  = FALSE
              )

# run time
print( runtime <- Sys.time() - start )


### Step 4: view results

smr <- summary( r )
print( smr )


### Step 5 (optional): parameter transformations and additional parameters
# extract chains
chains <- data.frame( rstan::extract( r$stanfit, permuted = FALSE, inc_warmup = TRUE, include = TRUE )[,1,] )
smpls.l <- list()
smpls.l$a                           <- chains[,"popmeans.1."]
smpls.l$sqrt.q                      <- chains[,"popmeans.2."]
smpls.l$b                           <- chains[,"popmeans.3."]
smpls.l$sqrt.sigma2b                <- chains[,"popsd.3."]
# (caution: names might differ in different models)

# discard burnin
burnin.share <- 0.10
smpls.l2 <- sapply( smpls.l, function( r ) { burnin.ind <- round( length( r )*burnin.share ) + 1;
                                             r[ ( burnin.ind ) : ( length( r ) ) ] }, simplify=FALSE )

# compute additional parameters
smpls.l2$q <- smpls.l2$sqrt.q^2
smpls.l2$sigma2mu <- 1/smpls.l2$a^2 * smpls.l2$sqrt.sigma2b^2
smpls.l2$mu <- -1/smpls.l2$a * smpls.l2$b

# function to compute the mode of the chain
fmode <- function( chain ){
   d <- density( chain )
   val <- d$x[ which( d$y  == max( d$y ) ) ]
   return( val )
}

## parameter estimates
( a <- fmode( smpls.l2$a ) )
( q <- fmode( smpls.l2$q ) )
( sigma2mu <- fmode( smpls.l2$sigma2mu ) )
( mu <- fmode( smpls.l2$mu ) )
