##########################################################################################################
#      R code for running three implementations of the univariate continuous-time model with JAGS        #
##########################################################################################################


## used R version: 3.6.1 (R Core Team, 2019)
# => get latest version: https://protect-us.mimecast.com/s/3yH6C2kg95ik6kB51t8WLZu?domain=cran.r-project.org/

## install JAGS from
# https://protect-us.mimecast.com/s/lbyTC4xk97tJ6JD4gCLTci8?domain=sourceforge.net
# used JAGS version: 4.3.0 (Plummer, 2017)

## load/install packages
# used rjags version: 4-9 (Plummer, 2019)
# => get latest rjags version: install.packages( "rjags" )
library( rjags )
# used shinystan version: 2.5.0 (Gabry, 2018)
# => get latest shinystan version: install.packages( "shinystan" )
library( shinystan )
# used truncnorm version: 1.0-8 (Mersmann, Trautmann, Steuer, & Bornkamp, 2018)
# => get latest truncnorm version: install.packages( "truncnorm" )
library( truncnorm )


## implementation to run
# select with corresponding number in [<no.>]
implementations <- c( "classic", "covbased", "covmeanbased" )
run <- implementations[2]

# bugs files
bugs.files <- c(
    "https://protect-us.mimecast.com/s/FLxOC5ylW5F0R0XkEhQszPo?domain=ndownloader.figshare.com", # bugs/jags code for the classic implementation
    "https://protect-us.mimecast.com/s/1WyIC68m05HoMopLzHEGfE0?domain=ndownloader.figshare.com", # bugs/jags code for the covariance-based implementation
    "https://protect-us.mimecast.com/s/u_TjC73nA5Imym8GlH7-0L5?domain=ndownloader.figshare.com"  # bugs/jags code for the covariance- and mean-based implementation
)
names( bugs.files ) <- implementations

# open bugs file for the implementation to run
bugs.file <- url( bugs.files[run] )

## approximate run times for this example:
# classic:      2 hrs 18 min
# covbased:     20 s
# covmeanbased: 40 s

## load simulated example data
## (simulated data with J=10000 persons, P=3 measurement occasions, a=-0.40, q=0.40, sigma2mu=0.50, sigma2fw=0.50, mu=0, mudev=0)
( load( url( "https://protect-us.mimecast.com/s/OkcQC31j95hmxmyDqHzu1f1?domain=ndownloader.figshare.com" ) ) )
# D: data matrix in wide format:
#    rows:     measurement occasions
#    columns:  persons
#    cells:    values of persons at measurement occasions
# delta: time interval length between measurement occasions

# number of persons
( J <- nrow( D ) )

# number of measurement occasions
( P <- ncol( D ) )

## data list for JAGS
data.list <- list()
data.list <- c( data.list, list( "D"=D ) )
data.list <- c( data.list, list( "delta"=delta ) )
data.list <- c( data.list, list( "J"=J ) )
data.list <- c( data.list, list( "P"=P ) )

## compute sample means and sample scatter matrix from data
# sample mean vector ybar
y.bar <- matrix( colMeans( D ), P, 1 )
# sample scatter matrix
S <- cov( D ) * ( J-1 )

# put S into and delete D and P from data.list for the reformulated implementations
if( run %in% c("covbased","covmeanbased") ){
   data.list$D <- NULL
   data.list$P <- NULL
   data.list$S <- S
   # additionally add sample means for the cov./mean-based implementation
   if( run %in% "covmeanbased" ){
      data.list$y.bar <- y.bar
   }
}

## JAGS run parameters
# thinning
thin <- 1
# maximum number of iterations
iter.max <- 200000
# block size (number of iterations after thinning per run)
iter.block.size <- 50
# maximum number of blocks (is calculated)
( n.iter.blocks <- round( iter.max / iter.block.size, 0 ) )
# burnin share
burnin.share <- 0.10
# stopping criteria for potential scale reduction factor (Rhat)
psr.crit <- 1.001
# stopping criteria for effective sample size
ess.crit <- 400 # set to higher values, e.g., 1000 for real data analysis
# include parameters when determining stopping
stop.include <- c('a','q','sigma2mu','sigma2fw','mu','mudev')
# parameters that are tracked
track <-        c('a','q','sigma2mu','sigma2fw','mu','mudev')
# delete 'mu' and 'mudev' for covariance-based implementation
if( run %in% "covbased" ){
   stop.include <- stop.include[ !stop.include %in% c('mu','mudev') ]
   track <- track[ !track %in% c('mu','mudev') ]
}

# function to generate random starting values
generate.startval <- function(){
   startval <- list()
   startval <- c( startval, list( "a"             = runif(1, -3, -1/10^5 ) ) )
   startval <- c( startval, list( "prec.q"        = 1/rtruncnorm(1, a=1/10^5, b=Inf, mean = 0, sd = sqrt(1) ) ) )
   startval <- c( startval, list( "prec.sigma2mu" = 1/rtruncnorm(1, a=1/10^5, b=Inf, mean = 0, sd = sqrt(1) ) ) )
   startval <- c( startval, list( "prec.sigma2fw" = 1/rtruncnorm(1, a=1/10^5, b=Inf, mean = 0, sd = sqrt(1) ) ) )
   startval <- c( startval, list( "mu"            = rnorm( 1, 0, sqrt(1) ) ) )
   startval <- c( startval, list( "mudev"         = rnorm( 1, 0, sqrt(1) ) ) )
   return( startval )
}

# generate starting values for the chain
set.seed( 128650203 ) # set seed for reproducible results
inits <- sapply( 1:1, function (chain.nr,l) l, generate.startval(), simplify=FALSE )
# delete starting values of 'mu' and 'mudev' for the covariance-based implementation
if( run %in% "covbased" ){
   inits[[1]]$mu <- NULL
   inits[[1]]$mudev <- NULL
}

# set seeds for the chain
# either use function parallel.seeds:
#   seeds <- parallel.seeds( 'base::BaseRNG', 1 )
# or set one specific seed for reproducible results
seeds <- list(list()); seeds[[1]]$.RNG.name <- "base::Super-Duper"; seeds[[1]]$.RNG.state <- c(-485017033, -2076628615)
inits <- mapply ( function(i,s) c(i,s), inits, seeds, SIMPLIFY=FALSE )

# function for obtaining parameter estimates: mode of the chain
fmode <- function( chain ){	d <- density( chain ); d$x[ which( d$y  == max( d$y ) ) ] }

# initializations for loop
continue <- TRUE # starts the loop
iter.block <- 1  # initial block number
smpls.l <- NULL  # object to collect samples

# start time
start <- Sys.time()

## loop until stopping criteria are reached
while( continue ){

   # print block number
   cat( paste0( "\n==================\nblock number: ", iter.block, "\n\n" ) ); flush.console()

   ## initialize of JAGS model
   # number of chains = 1 (code runs only for one chain)
   # number of adaptation iterations = 0 (no pre-sampling, just initialization)
   ini <- jags.model ( file=bugs.file, data=data.list, n.chains=1, n.adapt=0,
                       inits=inits, quiet=FALSE )

   # run JAGS model
   cat( paste0( "Sampling:", "\n" ) ); flush.console()
   res <- jags.samples ( ini, variable.names=track, n.iter=iter.block.size*thin, thin=thin,
                         type='trace', progress.bar='text', by=iter.block.size*thin/20 )

   # save RNG state
   RNGstate <- ini$state(internal=TRUE)[[1]]$".RNG.state"

   # cumulate samples
   if( is.null( smpls.l ) ){
      smpls.l <- sapply( res, function( r ) r[1,,1], simplify=FALSE )
   } else {
      smpls.l.cur <- sapply( res, function( r ) r[1,,1], simplify=FALSE )
      smpls.l <- mapply( function( l1, l2 ) c( l1, l2 ), smpls.l, smpls.l.cur, SIMPLIFY=FALSE )
   }

   # discard burnin
   smpls.l2 <- sapply( smpls.l, function( r ) { burnin.ind <- 
                       round( length( r )*burnin.share ) + 1;
                       r[ ( burnin.ind ) : ( length( r ) ) ] }, simplify=FALSE )

   # calculate potential scale reduction factor
   psr <- sapply( smpls.l2, function( chain ) { m<-matrix(chain,ncol=1); colnames(m)<-'x';
                  retrieve( as.shinystan( list( m ) ), 'rhat' ) } )
   # calculate effective sample size
   ess <- sapply( smpls.l2, function( chain ) { m<-matrix(chain,ncol=1); colnames(m)<-'x';
                  retrieve( as.shinystan( list( m ) ), 'neff' ) } )

   # print stats of current loop
   cat( "potential scale reduction factor:\n" ); print( psr )
   cat( "effective sample size:\n" ); print( ess )
   cat( "estimates:\n" ); print( est <- sapply( smpls.l2, fmode ) ); flush.console()

   # determine whether stopping criteria have been reached
   continue <- ( iter.block < n.iter.blocks ) & (( max ( psr[stop.include] )
                 > psr.crit ) | ( min ( ess[stop.include] ) < ess.crit ) )

   # prepare for next block
   if ( continue ){
      # set last value of the chain as starting value for the next block
      inits[[1]]$a <- smpls.l2$a[ length( smpls.l2$a ) ]
      inits[[1]]$prec.q <- 1 / smpls.l2$q[ length( smpls.l2$q ) ]
      inits[[1]]$prec.sigma2mu <- 1 / smpls.l2$sigma2mu[ length( smpls.l2$sigma2mu ) ]
      inits[[1]]$prec.sigma2fw <- 1 / smpls.l2$sigma2fw[ length( smpls.l2$sigma2fw ) ]
      if( !run %in% "covbased" ){
         inits[[1]]$mu <- smpls.l2$mu[ length( smpls.l2$mu ) ]
         inits[[1]]$mudev <- smpls.l2$mudev[ length( smpls.l2$mudev ) ]
      }

      # set RNG state on last state
      inits[[1]]$.RNG.state <- RNGstate
      
      # increase block number
      iter.block <- iter.block + 1
   }
}

# run time
( runtime <- Sys.time() - start )

# are stopping criteria for all parameters reached? (FALSE if iter.max reached without fulfilled stopping criteria)
all( (psr[stop.include] <= psr.crit) & ( ess[stop.include] >= ess.crit ) )

# print final psr, ess, and estimates
( psr )
( ess )
( est )

# close bugs file
close( bugs.file )
